# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_FactoryAuth.ipynb.

# %% auto 0
__all__ = ['generate_config_auth', 'Config_GenerateConfigAuth', 'upsert_domoaccount', 'generate_config_access_token',
           'Config_UpsertDomoAccessTokenAccount']

# %% ../nbs/00_FactoryAuth.ipynb 1
from .utils.factory import (
    factory_function,
    FactoryLogs, 
    FactoryResponse, FactoryConfig, FactoryMessage)

# %% ../nbs/00_FactoryAuth.ipynb 2
from dataclasses import dataclass
from typing import Callable

import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoAccount as dmac
import domolibrary_execution.utils.domojupyter as dxdj

import httpx

# %% ../nbs/00_FactoryAuth.ipynb 5
@factory_function
async def generate_config_auth(
    config : FactoryConfig,
    res: FactoryResponse,
    
    config_auth_standard : dmda.DomoAuth,
    config_auth_exception : dmda.DomoAuth,
    target_instance_use_exception_pw : bool,
    target_instance : str,

    # ouptut
    auth : dmda.DomoAuth =None, # will be updated during execution
    
    **kwargs
):

    res.location = target_instance
    step = FactoryMessage(stage="generate_config_auth", message="using standard auth")
    res.add_message(step)
    auth = config_auth_standard
    step.is_success = True

    if target_instance_use_exception_pw == 1:
        auth = config_auth_exception
        step.message = "using exception auth"
        step.is_success = True

    auth.domo_instance = target_instance

    step = FactoryMessage(stage="test is valid auth")
    res.add_message(step)

    try:
        await auth.print_is_token()
        step.is_success = True
        step.message = "valid auth"

    except Exception as e:
        step.message = e
        step.is_success = False

    config.auth = auth
    res.response = auth
    return res

@dataclass
class Config_GenerateConfigAuth(FactoryConfig):
    config_auth_exception : dmda.DomoAuth
    config_auth_standard : dmda.DomoAuth
    target_instance : str
    target_instance_use_exception_pw : bool = True

    # updated during execution
    auth: dmda.DomoAuth = None

# %% ../nbs/00_FactoryAuth.ipynb 8
@factory_function
async def upsert_domoaccount(
    config: FactoryConfig,
    res : FactoryResponse,

    auth : dmda.DomoAuth,
    
    account_name : str,
    account_config_cls : str,
    generate_config_fn : Callable,

    modify_config_fn: Callable = None,
    session : httpx.AsyncClient = None,
    debug_api: bool = False,

    # used by wrapper
    **kwargs

):
    res.location = auth.domo_instance

    step = FactoryMessage( stage="upsert domo account", is_success=False)
    res.add_message( step )
    
    
    domo_account = None
    try:
        account_config = generate_config_fn(config, account_config_cls)
    
        domo_account = await dmac.DomoAccounts.upsert_account(
            auth = auth,
            account_config = account_config,
            account_name = account_name,
            session = session,
            debug_api = debug_api
        )

        config.domo_account = domo_account
    
        step.is_success = True
        step.message = f'account {account_name} upserted'
        
    except Exception as e:
        step.message = e
        step.is_success = False
        

    if modify_config_fn:
        step = FactoryMessage( stage="modify config", is_success=False)
        res.add_message( step )
    
        modify_config_fn(config = config, step = step, 
                         domo_account = domo_account)
        
           
    res.response = domo_account or 'no account'
        
    return res



# for different accounts  config object will need different fields
def generate_config_access_token(config, account_config_cls : dmac.AccountConfig.domo_access_token):
    return account_config_cls(
        domo_access_token = config.access_token,
        username = config.username,
        password = config.password
    )

@dataclass
class Config_UpsertDomoAccessTokenAccount(FactoryConfig):
    account_name : str 
    generate_config_fn : Callable
    account_config_cls : dmac.AccountConfig.domo_access_token.value

    ## account params
    username : str
    access_token : str = None
    password : str = None

    # updated during execution
    domo_account : dmac.DomoAccount = None





# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/factory/factory.ipynb.

# %% auto 0
__all__ = ['FactoryMessage', 'FactoryResponse', 'FactoryLogs', 'FactoryFunction_MissingParameter', 'FactoryConfig',
           'FactoryFunction_ResponseTypeError', 'FactoryFunction_Error', 'FactoryFunction_ResponseError',
           'FactoryFunction_NotSuccess', 'factory_function']

# %% ../../nbs/factory/factory.ipynb 2
from dataclasses import dataclass, field
from typing import Any, Callable, List
import functools

import httpx

import domolibrary_extensions.client as dec

# %% ../../nbs/factory/factory.ipynb 6
@dataclass
class FactoryMessage:
    stage: str  # description of the stage of a process
    message: str = "init"  # outcome
    stage_num: int = None
    is_success: bool = False

    """class for logging a stage of a process"""

    def to_json(self):
        return {key: value for key, value in self.__dict__.items() if value is not None}

    def to_string(self):
        string = " | ".join(
            [
                f"{key} - {str(value)}"
                for key, value in self.__dict__.items()
                if value is not None
            ]
        )

        if not self.is_success:
            string = f"💀 {string}"
        return string

    def __eq__(self, other):
        if not isinstance(other, FactoryMessage):
            return False
        return self.stage == other.stage

# %% ../../nbs/factory/factory.ipynb 10
@dataclass
class FactoryResponse:
    function_name: str
    id: str  # identify a set of log entries

    messages: List[FactoryMessage] = field(
        default_factory=lambda: []
    )  # capture intermediate steps of the factory_function
    response: Any = field(
        repr=False, default=None
    )  # final object to return from the factory_function
    is_success: bool = False
    location: str = None

    """Response class for handling logging of a factory function.
    Accumulates messages as the factory unfolds
    """

    def to_json(self) -> List[dict]:
        columns = [
            "location",
            "function_name",
            "stage_num",
            "stage",
            "id",
            "is_success",
            "message",
            "response",
        ]
        s = [{**self.__dict__, **msg.to_json()} for msg in self.messages]

        return [{col: obj[col] for col in columns if obj.get(col)} for obj in s]

    def add_message(self, message: FactoryMessage):
        if message in self.messages:
            print("message in messages", message)

        self.messages.append(message)

    def test_success(self):
        """tests if all factory_messages are successful"""
        self.is_success = all([message.is_success for message in self.messages])
        return self.is_success

    def __eq__(self, other):
        if not isinstance(other, FactoryResponse):
            return False

        return self.id == other.id and self.function_name == other.function_name

# %% ../../nbs/factory/factory.ipynb 14
@dataclass
class FactoryLogs:
    """factory logs are the complete logs of an entire factory or script"""

    logs: List[FactoryResponse] = field(default_factory=lambda: [])

    def add_response(self, res: FactoryResponse):
        if res not in self.logs:
            self.logs.append(res)

        return self.logs

    def to_json(self) -> List[dict]:
        return [message for log in self.logs for message in log.to_json()]

# %% ../../nbs/factory/factory.ipynb 18
class FactoryFunction_MissingParameter(Exception):
    def __init__(self, parameter, function_name):

        super().__init__(
            f"missing parameter - {parameter} - while calling {function_name}"
        )

@dataclass
class FactoryConfig:
    """create a dataclass which will have all the fields used in your config event"""

    logs: FactoryLogs
    config_id_col: str
    location: str
    factory_fn_ls: List[Callable]

    session: httpx.AsyncClient = field(repr=False)

    async def run(self, factory_fn_ls=None, debug_api: bool = False, debug_run : bool = False ):

        factory_fn_ls = factory_fn_ls or self.factory_fn_ls

        for fn in factory_fn_ls:

            if debug_run:
                print(f"running - {fn.__name__}")
    
            await fn(config=self, debug_api=debug_api, fn_name = fn.__name__, **self.__dict__)

        return self

# %% ../../nbs/factory/factory.ipynb 19
class FactoryFunction_ResponseTypeError(TypeError):
    """a function wrapped in `process_function` must return FactoryResponse class"""

    def __init__(self, result):
        super().__init__(
            f"Expected function to return an instance of FactoryResponse. Got {type(result)} instead.  Refactor function to return FactoryResponse class"
        )


class FactoryFunction_Error(Exception):
    """base class for capturing errors within a factory function"""

    def __init__(
        self, factory: FactoryResponse, message: FactoryMessage, location: str = None
    ):

        e = f"💀 | {factory.function_name } | {factory.id} | {message.stage} - {message.message}"

        if location:
            e = f"{e} | in {location}"

        super().__init__(e)


class FactoryFunction_ResponseError(Exception):
    def __init__(self, message):
        super().__init__(message)


class FactoryFunction_NotSuccess(Exception):
    def __init__(self, messages: List[FactoryMessage]):
        super().__init__(
            "\n".join(
                [message.to_string() for message in messages if not message.is_success]
            )
        )

# %% ../../nbs/factory/factory.ipynb 20
def factory_function(func: Callable[..., Any]) -> Callable[..., Any]:

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):

        logs = kwargs.pop("logs")
        config_id_col = kwargs.pop("config_id_col")
        location = kwargs.pop("location")
        config = kwargs.pop("config")
        func.__name__ = kwargs.pop('fn_name')

        for req in [logs, config_id_col, location]:
            if req is None:
                raise FactoryFunction_MissingParameter(req.__name__, func.__name__)

        id = getattr( config, config_id_col)

        res = FactoryResponse(
            function_name=func.__name__,
            id=id,
            location=location,
        )
        logs.add_response(res)

        result = await func(
            *args,
            config = config,
            res=res,
            logs=logs,
            location=location,
            **kwargs,
        )

        if not isinstance(result, FactoryResponse):
            raise FactoryFunction_ResponseTypeError(result)

        if not result.response:
            raise FactoryFunction_ResponseError(
                "factory response must return res.response"
            )

        if not result.messages or len(result.messages) == 0:
            raise FactoryFunction_ResponseError("factory response must have messages")

        result.test_success()

        if not result.is_success:
            raise FactoryFunction_NotSuccess(result.messages)

        return result

    return wrapper



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_FactoryUser.ipynb.

# %% auto 0
__all__ = ['upsert_user', 'UpsertUserConfig']

# %% ../nbs/00_FactoryUser.ipynb 2
from .utils.factory import (
    factory_function,
    FactoryLogs, 
    FactoryResponse, FactoryConfig, FactoryMessage, FactoryFunction_Error)

# %% ../nbs/00_FactoryUser.ipynb 3
from typing import List, Callable
from dataclasses import dataclass, field
import httpx

import domolibrary.client.DomoAuth as dmda

import domolibrary.classes.DomoInstanceConfig as dmic
import domolibrary.classes.DomoRole as dmdr
import domolibrary.classes.DomoUser as dmdu
import domolibrary.classes.DomoGroup as dmgr
import domolibrary.classes.DomoAccount as dmac 


# %% ../nbs/00_FactoryUser.ipynb 6
@factory_function
async def upsert_user(
    # base required
    logs: FactoryLogs,
    res : FactoryResponse,
    config : FactoryConfig,
    auth: dmda.DomoAuth, 
    
    email : str,
    display_name : str,
    domo_role : dmdr.DomoRole = None, # will use default role if no role provided
    password: str = None,
    
    # OUTPUT
    domo_user : dmdu.DomoUser = None, #will add domo user to config

    # DEBUG
    debug_api: bool = False,
    session : httpx.AsyncClient = None,
    **kwargs
    ):
    
    res.response = False    
    auth = auth
    
    # STAGE_1 = DOMO ROLE OR DEFAULT ROLE
    step = FactoryMessage( stage="retrieve domo role", is_success=False)
    res.add_message( step )
    
    if not domo_role:
        try:
            domo_role = await dmdr.DomoRoles.get_default_role(auth = auth,debug_api = debug_api, session = session)
            config.domo_role = domo_role
            
        except Exception as e:
            step.is_success = False
            step.message = str(e)

    if not domo_role:
        res.response = "no domo role"
        return res
    
    step.message = f"using role {domo_role.id} - {domo_role.name}"
    step.is_success = True
            
    # STAGE_2 = DOMO_USER
    step = FactoryMessage( stage="upsert user", is_success=False)
    res.add_message( step )
    
    try:
        domo_user  = await dmdu.DomoUsers.upsert_user(
            auth = auth,
            email_address = email,
            display_name = display_name,
            session = session,
            role_id = domo_role.id)      
                      
        config.domo_user = domo_user
        step.message = f"upsert user {domo_user.id} successful"
        step.is_success = True
    
    except Exception as e:
        step.message = e
        step.is_success = False
    
    if not domo_user:
        res.response = "no domo user"
        return res

    res.response = domo_user
    
    if not password:
        return res
    
    # STEP_3 = SET PASSWORD
    step = FactoryMessage( stage="set password", is_success=False)
    res.add_message( step )
        
    try:
        await domo_user.reset_password(new_password = password, debug_api = debug_api)
        domo_user.password = password
        step.message = f"set password {password}"
        step.is_success = True
    
    except Exception as e:
        step.message = e
        step.is_success = False
    
    return res

@dataclass
class UpsertUserConfig(FactoryConfig):
    auth: dmda.DomoAuth
    display_name : str
    email: str
    password :str = None # will not set password if not provided
    domo_role : dmdr.DomoRole = None # will use default not passed
    domo_user : dmdu.DomoUser = None # will assign after upsert

    # factory_fn_ls : List[Callable] = field(default_factory = lambda:[sync_user])

